# stdlib
from typing import Any
from typing import cast

# third party
from IPython.display import display
from rich.prompt import Confirm
from typing_extensions import Self

# relative
from ..abstract_server import AbstractServer
from ..abstract_server import ServerSideType
from ..abstract_server import ServerType
from ..serde.serializable import serializable
from ..server.credentials import SyftCredentials
from ..types.base import SyftBaseModel
from ..types.syft_object import Context
from .user.user_roles import ServiceRole


class WarningContext(
    Context,
):
    server: AbstractServer | None = None
    credentials: SyftCredentials | None = None
    role: ServiceRole


@serializable(canonical_name="APIEndpointWarning", version=1)
class APIEndpointWarning(SyftBaseModel):
    confirmation: bool = False
    message: str | None = None
    enabled: bool = True

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, APIEndpointWarning):
            return self.message == other.message and self._bool == other._bool
        return self._bool == other

    @property
    def _repr_html_class_(self) -> str:
        return "alert-warning"

    def __repr__(self) -> str:
        return f"SyftWarning: {self.message}"

    def __str__(self) -> str:
        return self.__repr__()

    def _repr_html_(self) -> str:
        return (
            f'<div class="{self._repr_html_class_}" style="padding:5px;">'
            + f"<strong>SyftWarning</strong>: {self.message}</div><br />"
        )

    def message_from(self, context: WarningContext | None) -> Self:
        raise NotImplementedError

    def show(self) -> bool:
        if not self.enabled or not self.message:
            return True
        display(self)
        if self.confirmation:
            allowed = Confirm.ask("Would you like to proceed?")
            if not allowed:
                display("Aborted !!")
                return False
        return True


@serializable(canonical_name="CRUDWarning", version=1)
class CRUDWarning(APIEndpointWarning):
    def message_from(self, context: WarningContext | None = None) -> Self:
        message = None
        confirmation = self.confirmation
        if context is not None:
            server = context.server
            if server is not None:
                server_side_type = cast(ServerSideType, server.server_side_type)
                server_type = server.server_type

                _msg = (
                    "which could host datasets with private information."
                    if server_side_type.value == ServerSideType.HIGH_SIDE.value
                    else "which only hosts mock or synthetic data."
                )
                if server_type is not None:
                    message = (
                        "You're performing an operation on "
                        f"{server_side_type.value} side {server_type.value}, {_msg}"
                    )
                confirmation = server_side_type.value == ServerSideType.HIGH_SIDE.value

        return CRUDWarning(confirmation=confirmation, message=message)


@serializable(canonical_name="CRUDReminder", version=1)
class CRUDReminder(CRUDWarning):
    confirmation: bool = False

    def message_from(self, context: WarningContext | None = None) -> Self:
        message = None
        confirmation = self.confirmation
        if context is not None:
            server = context.server
            if server is not None:
                server_side_type = cast(ServerSideType, server.server_side_type)
                server_type = server.server_type

                _msg = (
                    "which could host datasets with private information."
                    if server_side_type.value == ServerSideType.HIGH_SIDE.value
                    else "which only hosts mock or synthetic data."
                )
                if server_type is not None:
                    message = (
                        "You're performing an operation on "
                        f"{server_side_type.value} side {server_type.value}, {_msg}"
                    )

        return CRUDReminder(confirmation=confirmation, message=message)


@serializable(canonical_name="LowSideCRUDWarning", version=1)
class LowSideCRUDWarning(APIEndpointWarning):
    def message_from(self, context: WarningContext | None = None) -> Self:
        confirmation = self.confirmation
        message = None
        if context is not None:
            server = context.server
            if server is not None:
                server_side_type = cast(ServerSideType, server.server_side_type)
                server_type = cast(ServerType, server.server_type)
                if server_side_type.value == ServerSideType.LOW_SIDE.value:
                    message = (
                        "You're performing an operation on "
                        f"{server_side_type.value} side {server_type.value} "
                        "which only hosts mock or synthetic data."
                    )

        return LowSideCRUDWarning(confirmation=confirmation, message=message)


@serializable(canonical_name="HighSideCRUDWarning", version=1)
class HighSideCRUDWarning(APIEndpointWarning):
    def message_from(self, context: WarningContext | None = None) -> Self:
        confirmation = self.confirmation
        message = None
        if context is not None:
            server = context.server
            if server is not None:
                server_side_type = cast(ServerSideType, server.server_side_type)
                server_type = cast(ServerType, server.server_type)
                if server_side_type.value == ServerSideType.HIGH_SIDE.value:
                    message = (
                        "You're performing an operation on "
                        f"{server_side_type.value} side {server_type.value} "
                        "which could host datasets with private information."
                    )

        return HighSideCRUDWarning(confirmation=confirmation, message=message)



# stdlib

# third party
from result import Err
from result import Ok
from result import Result

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.document_store import BaseUIDStoreStash
from ...store.document_store import PartitionKey
from ...store.document_store import PartitionSettings
from ...store.document_store import QueryKeys
from ...store.linked_obj import LinkedObject
from ...types.datetime import DateTime
from ...types.uid import UID
from ...util.telemetry import instrument
from .notifications import Notification
from .notifications import NotificationStatus

FromUserVerifyKeyPartitionKey = PartitionKey(
    key="from_user_verify_key", type_=SyftVerifyKey
)
ToUserVerifyKeyPartitionKey = PartitionKey(
    key="to_user_verify_key", type_=SyftVerifyKey
)
StatusPartitionKey = PartitionKey(key="status", type_=NotificationStatus)

OrderByCreatedAtTimeStampPartitionKey = PartitionKey(key="created_at", type_=DateTime)

LinkedObjectPartitionKey = PartitionKey(key="linked_obj", type_=LinkedObject)


@instrument
@serializable(canonical_name="NotificationStash", version=1)
class NotificationStash(BaseUIDStoreStash):
    object_type = Notification
    settings: PartitionSettings = PartitionSettings(
        name=Notification.__canonical_name__,
        object_type=Notification,
    )

    def get_all_inbox_for_verify_key(
        self, credentials: SyftVerifyKey, verify_key: SyftVerifyKey
    ) -> Result[list[Notification], str]:
        qks = QueryKeys(
            qks=[
                ToUserVerifyKeyPartitionKey.with_obj(verify_key),
            ]
        )
        return self.get_all_for_verify_key(
            credentials=credentials, verify_key=verify_key, qks=qks
        )

    def get_all_sent_for_verify_key(
        self, credentials: SyftVerifyKey, verify_key: SyftVerifyKey
    ) -> Result[list[Notification], str]:
        qks = QueryKeys(
            qks=[
                FromUserVerifyKeyPartitionKey.with_obj(verify_key),
            ]
        )
        return self.get_all_for_verify_key(credentials, verify_key=verify_key, qks=qks)

    def get_all_for_verify_key(
        self, credentials: SyftVerifyKey, verify_key: SyftVerifyKey, qks: QueryKeys
    ) -> Result[list[Notification], str]:
        if isinstance(verify_key, str):
            verify_key = SyftVerifyKey.from_string(verify_key)
        return self.query_all(
            credentials,
            qks=qks,
            order_by=OrderByCreatedAtTimeStampPartitionKey,
        )

    def get_all_by_verify_key_for_status(
        self,
        credentials: SyftVerifyKey,
        verify_key: SyftVerifyKey,
        status: NotificationStatus,
    ) -> Result[list[Notification], str]:
        qks = QueryKeys(
            qks=[
                ToUserVerifyKeyPartitionKey.with_obj(verify_key),
                StatusPartitionKey.with_obj(status),
            ]
        )
        return self.query_all(
            credentials,
            qks=qks,
            order_by=OrderByCreatedAtTimeStampPartitionKey,
        )

    def get_notification_for_linked_obj(
        self,
        credentials: SyftVerifyKey,
        linked_obj: LinkedObject,
    ) -> Result[Notification, str]:
        qks = QueryKeys(
            qks=[
                LinkedObjectPartitionKey.with_obj(linked_obj),
            ]
        )
        return self.query_one(credentials=credentials, qks=qks)

    def update_notification_status(
        self, credentials: SyftVerifyKey, uid: UID, status: NotificationStatus
    ) -> Result[Notification, str]:
        result = self.get_by_uid(credentials, uid=uid)
        if result.is_err():
            return result.err()

        notification = result.ok()
        if notification is None:
            return Err(f"No notification exists for id: {uid}")
        notification.status = status
        return self.update(credentials, obj=notification)

    def delete_all_for_verify_key(
        self, credentials: SyftVerifyKey, verify_key: SyftVerifyKey
    ) -> Result[bool, str]:
        result = self.get_all_inbox_for_verify_key(
            credentials,
            verify_key=verify_key,
        )
        # If result is an error then return the error
        if result.is_err():
            return result

        # get the list of notifications
        notifications = result.ok()

        for notification in notifications:
            result = self.delete_by_uid(credentials, uid=notification.id)
            if result.is_err():
                return result
        return Ok(True)
